from datetime import datetime, tzinfo
from typing import Type, Any, TypeVar, Tuple, List

from data_aggregator_sdk.integration_message import IntegrationV0MessageData
from pydantic import BaseModel, PrivateAttr
from ul_api_utils.conf import APPLICATION_F

from data_gateway_sdk.errors import DataGatewayDeviceProtocolParsingError
from data_gateway_sdk.utils.packet import Packet

T = TypeVar('T')


class DevicePacket(BaseModel):
    @classmethod
    def parse(cls: Type[T], payload: bytes, **kwargs: Any) -> List[T]:
        raise NotImplementedError()

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        raise NotImplementedError()


TClass = TypeVar('TClass', bound='DevicePacketTyped')


class DevicePacketTyped(BaseModel):
    packet: Packet
    _SUPPORTED_PACKETS: Tuple[Type[Packet], ...] = PrivateAttr(default=())

    @classmethod
    def parse(cls: Type[TClass], payload: bytes, **kwargs: Any) -> List[TClass]:
        supported_packets = cls.__private_attributes__["_SUPPORTED_PACKETS"].default    # pydanticV2 migration https://docs.pydantic.dev/latest/concepts/models/#private-model-attributes
        if not APPLICATION_F.has_or_unset(cls.__name__):
            raise DataGatewayDeviceProtocolParsingError(f'payload "{payload.hex()}" not matched with ({", ".join(c.__name__ for c in supported_packets)})')  # type: ignore
        parsed_packets = Packet.parse_packets(payload, supported_packets)  # type: ignore
        if not parsed_packets:
            raise DataGatewayDeviceProtocolParsingError(
                f'payload "{payload.hex()}" not matched with ({", ".join(c.__name__ for c in supported_packets)})')  # type: ignore

        return [cls(packet=parsed_packet) for parsed_packet in parsed_packets]

    def to_integration_data(self, *, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return self.packet.to_integration_data(received_at=received_at, device_tz=device_tz, **kwargs)


class EmptyDevicePacket(DevicePacket):
    @classmethod
    def parse(cls: Type[T], payload: bytes, **kwargs: Any) -> List[T]:
        raise TypeError(f'{cls.__name__} not supported parse message')

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
            ),
        ]



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h_f import SmpmUlDeviceEnergy16BProfile8HFData, SmpmUlDeviceEnergy16bProfile8hFIds
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_profile_8h_f() -> None:
    case_serialized = bytes.fromhex("b40ee001144010fa10d482fcb80b0000")
    assert SmpmUlDeviceEnergy16BProfile8HFData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG, days_ago=timedelta(seconds=0.0), profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)) == SmpmUlDeviceEnergy16BProfile8HFData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HFData.serialize(SmpmUlDeviceEnergy16BProfile8HFData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG, days_ago=timedelta(seconds=0.0), profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)))  # noqa: E501
    case_serialized = bytes.fromhex("b40ee001144010fa10d482fcb80b0000")
    assert SmpmUlDeviceEnergy16BProfile8HFData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG, days_ago=timedelta(seconds=0.0), profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)) == SmpmUlDeviceEnergy16BProfile8HFData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HFData.serialize(SmpmUlDeviceEnergy16BProfile8HFData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG, profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)))  # noqa: E501
    case_serialized = bytes.fromhex("d70ee001144010fa10d482fcb80b0000")
    assert SmpmUlDeviceEnergy16BProfile8HFData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_C, days_ago=timedelta(seconds=0.0), profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)) == SmpmUlDeviceEnergy16BProfile8HFData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HFData.serialize(SmpmUlDeviceEnergy16BProfile8HFData(days_ago=timedelta(seconds=0.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_C, profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)))  # noqa: E501
    case_serialized = bytes.fromhex("b4ceef01144010fa10d482fcb80b0000")
    assert SmpmUlDeviceEnergy16BProfile8HFData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG, days_ago=timedelta(seconds=5443200.0), profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)) == SmpmUlDeviceEnergy16BProfile8HFData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HFData.serialize(SmpmUlDeviceEnergy16BProfile8HFData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_1_F_AVG, profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)))  # noqa: E501
    case_serialized = bytes.fromhex("d7ceef01144010fa10d482fcb80b0000")
    assert SmpmUlDeviceEnergy16BProfile8HFData(packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_C, days_ago=timedelta(seconds=5443200.0), profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)) == SmpmUlDeviceEnergy16BProfile8HFData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8HFData.serialize(SmpmUlDeviceEnergy16BProfile8HFData(days_ago=timedelta(seconds=5443200.0), packet_type_id=SmpmUlDeviceEnergy16bProfile8hFIds.UL_DATA_16B__PROFILE_H8_3_F_MAX__PHASE_C, profile=(45.3, 45.4, 55.4, 65.0, 55.4, 45.9, 55.1, 60.0)))  # noqa: E501



from decimal import Decimal
from typing import Dict, List, Any
from datetime import datetime, tzinfo

from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageConsumption, \
    CounterType, ResourceType

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   water5_device_info_ch2
#
# RESULT int:        4172905
# RESULT bin:  MSB   0000000000000000000000000000000000000000001111111010110001101001   LSB
# RESULT hex:  LE    69 ac 3f 00 00 00 00 00
#
#
# name      type    size  value(int)                                                        data(bits)
# ----------------------------------------------------------------------------------------------------
# pack_id   u8         8         105                                                          01101001
# value     uf32p3    32       16300                          00000000000000000011111110101100
# RESERVED  u24       24           0  000000000000000000000000


class Water5DeviceInfoCh2Data(Packet):
    value: float

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((105) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.value, (int, float))
        assert 0.0 <= data.value <= 4294967.295
        result |= ((int(round(float(data.value) * 1000.0, 0))) & (2 ** (32) - 1)) << size
        size += 32
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'Water5DeviceInfoCh2Data':
        result__el_tmp1: Dict[str, Any] = dict()
        if 105 != buf.shift(8):
            raise ValueError("pack_id: buffer doesn't match value")
        result__el_tmp1["value"] = round(buf.shift(32) / 1000.0 - 0.0, 3)
        buf.shift(24)
        result = Water5DeviceInfoCh2Data(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=2,
                        value=Decimal(str(self.value)),
                        overloading_value=None,
                    ),
                    # : List[IntegrationV0MessageConsumption]
                ],
            ),
        ]



from decimal import Decimal
from datetime import datetime, tzinfo, time
from enum import IntEnum, unique
from typing import Dict, Any, List

from data_aggregator_sdk.constants.enums import JournalDataType
from data_aggregator_sdk.integration_message import IntegrationV0MessageConsumption, IntegrationV0MessageGeneration, IntegrationV0MessageData, \
    CounterType, ResourceType
from dateutil.relativedelta import relativedelta

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.first_day_of_month import first_day_of_month
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_retrospective_energy
#
# RESULT int:        60967556582773559757215997875516304
# RESULT bin:  MSB   00000000 00001011 10111101 11101110 11000000 00000000 00000000 00000000 10000100 00100111 01000011 01100111 10001110 00011000 11110011 10010000   LSB
# RESULT hex:  LE    90 f3 18 8e 67 43 27 84 00 00 00 c0 ee bd 0b 00
#
# name                       type    size  value(int)                                                                                                                        data(bits) # noqa: E501
# -------------------------  ------  ----  ----------  -------------------------------------------------------------------------------------------------------------------------------- # noqa: E501
# packet_type_id_enum.0.VAL  u7         7          16                                                                                                                           0010000 # noqa: E501
# packet_type_id_enum.0.DFF  bool       1           1                                                                                                                          1
# packet_type_id_enum.1.VAL  u2         2           3                                                                                                                        11
# packet_type_id_enum.1.DFF  bool       1           0                                                                                                                       0
# is_valid                   bool       1           0                                                                                                                      0
# period_ago                 u5         5          15                                                                                                                 01111
# value_current              uf27p2    27    28559116                                                                                      001101100111100011100001100
# value_previous_1_delta     uf24p2    24      541300                                                              000010000100001001110100
# value_previous_2_delta     uf24p2    24           0                                      000000000000000000000000
# value_previous_3_delta     uf24p2    24    12312300              101110111101111011101100
# RESERVED                   u12       12           0  000000000000

@unique
class SmpmUlDeviceEnergy16bRetrospectiveEnergyIds(IntEnum):
    DAILY_ENERGY_ACTIVE_CONSUMED = 400
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_1 = 401
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_2 = 402
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_3 = 403
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_4 = 404
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM = 405
    DAILY_ENERGY_REACTIVE_CONSUMED = 406
    DAILY_ENERGY_ACTIVE_GENERATED = 407
    DAILY_ENERGY_REACTIVE_GENERATED = 408
    MONTHLY_ENERGY_ACTIVE_CONSUMED = 409
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_1 = 410
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_2 = 411
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_3 = 412
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_4 = 413
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM = 414
    MONTHLY_ENERGY_REACTIVE_CONSUMED = 415
    MONTHLY_ENERGY_ACTIVE_GENERATED = 416
    MONTHLY_ENERGY_REACTIVE_GENERATED = 417


class SmpmUlDeviceEnergy16BRetrospectiveEnergyData(Packet):
    packet_type_id_enum: SmpmUlDeviceEnergy16bRetrospectiveEnergyIds
    is_valid: bool
    period_ago: int
    value_current: float
    value_previous_1_delta: float
    value_previous_2_delta: float
    value_previous_3_delta: float

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        packet_type_id_enum__value_int_tmp1 = 0
        packet_type_id_enum__value_size_tmp2 = 0
        assert isinstance(data.packet_type_id_enum, SmpmUlDeviceEnergy16bRetrospectiveEnergyIds)
        packet_type_id_enum__value_int_tmp1 |= ((data.packet_type_id_enum.value) & (2 ** (11) - 1)) << packet_type_id_enum__value_size_tmp2
        packet_type_id_enum__value_size_tmp2 += 11
        packet_type_id_enum__max_size_tmp3 = 0
        packet_type_id_enum__steps_tmp4 = [8, 3, 3, 3]
        for packet_type_id_enum__j_tmp5 in range(32):
            packet_type_id_enum__step_tmp6 = packet_type_id_enum__steps_tmp4[packet_type_id_enum__j_tmp5] if packet_type_id_enum__j_tmp5 < len(packet_type_id_enum__steps_tmp4) else packet_type_id_enum__steps_tmp4[-1]
            packet_type_id_enum__max_size_tmp3 += packet_type_id_enum__step_tmp6
            packet_type_id_enum__current_part_value_tmp7 = packet_type_id_enum__value_int_tmp1 & (2 ** packet_type_id_enum__step_tmp6 - 1)
            packet_type_id_enum__value_int_tmp1 = packet_type_id_enum__value_int_tmp1 >> (packet_type_id_enum__step_tmp6 - 1)
            result |= ((packet_type_id_enum__current_part_value_tmp7) & (2 ** ((packet_type_id_enum__step_tmp6 - 1)) - 1)) << size
            size += (packet_type_id_enum__step_tmp6 - 1)
            assert isinstance((packet_type_id_enum__value_int_tmp1 != 0), bool)
            result |= ((int((packet_type_id_enum__value_int_tmp1 != 0))) & (2 ** (1) - 1)) << size
            size += 1
            if packet_type_id_enum__value_int_tmp1 == 0:
                break
        assert isinstance(data.is_valid, bool)
        result |= ((int(data.is_valid)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.period_ago, int)
        assert 0 <= data.period_ago <= 31
        result |= ((data.period_ago) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.value_current, (int, float))
        result |= ((int(round(float(data.value_current) * 100.0, 0)) & 134217727) & (2 ** (27) - 1)) << size
        size += 27
        assert isinstance(data.value_previous_1_delta, (int, float))
        result |= ((int(round(float(data.value_previous_1_delta) * 100.0, 0)) & 16777215) & (2 ** (24) - 1)) << size
        size += 24
        assert isinstance(data.value_previous_2_delta, (int, float))
        result |= ((int(round(float(data.value_previous_2_delta) * 100.0, 0)) & 16777215) & (2 ** (24) - 1)) << size
        size += 24
        assert isinstance(data.value_previous_3_delta, (int, float))
        result |= ((int(round(float(data.value_previous_3_delta) * 100.0, 0)) & 16777215) & (2 ** (24) - 1)) << size
        size += 24
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BRetrospectiveEnergyData':
        result__el_tmp8: Dict[str, Any] = dict()
        packet_type_id_enum__res_tmp9 = 0
        packet_type_id_enum__steps_tmp10 = (8, 3, 3, 3)
        packet_type_id_enum__res_size_tmp11 = 0
        packet_type_id_enum__step_tmp12 = 0
        for packet_type_id_enum__i_tmp14 in range(32):
            packet_type_id_enum__step_tmp12 = (packet_type_id_enum__steps_tmp10[packet_type_id_enum__i_tmp14] if packet_type_id_enum__i_tmp14 < len(
                packet_type_id_enum__steps_tmp10) else packet_type_id_enum__steps_tmp10[-1]) - 1
            packet_type_id_enum__res_tmp9 |= buf.shift(packet_type_id_enum__step_tmp12) << packet_type_id_enum__res_size_tmp11
            packet_type_id_enum__res_size_tmp11 += packet_type_id_enum__step_tmp12
            packet_type_id_enum__dff_tmp13 = bool(buf.shift(1))
            if not packet_type_id_enum__dff_tmp13:
                break
        packet_type_id_enum__buf_tmp15 = buf
        buf = BufRef(packet_type_id_enum__res_tmp9, stop_on_buffer_end=True)
        result__el_tmp8["packet_type_id_enum"] = SmpmUlDeviceEnergy16bRetrospectiveEnergyIds(buf.shift(11))
        buf = packet_type_id_enum__buf_tmp15
        result__el_tmp8["is_valid"] = bool(buf.shift(1))
        result__el_tmp8["period_ago"] = buf.shift(5) + 0
        result__el_tmp8["value_current"] = round(buf.shift(27) / 100.0, 2)
        result__el_tmp8["value_previous_1_delta"] = round(buf.shift(24) / 100.0, 2)
        result__el_tmp8["value_previous_2_delta"] = round(buf.shift(24) / 100.0, 2)
        result__el_tmp8["value_previous_3_delta"] = round(buf.shift(24) / 100.0, 2)
        result = SmpmUlDeviceEnergy16BRetrospectiveEnergyData(**result__el_tmp8)
        buf.shift(12)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        resource_type: ResourceType
        tariff: int
        journal_data_type_for_end: JournalDataType
        period_ago: relativedelta

        if self.packet_type_id_enum in {401, 410}:
            tariff = 1
        elif self.packet_type_id_enum in {402, 411}:
            tariff = 2
        elif self.packet_type_id_enum in {403, 412}:
            tariff = 3
        elif self.packet_type_id_enum in {404, 413}:
            tariff = 4
        elif self.packet_type_id_enum in {405, 414}:
            tariff = 0
        else:
            tariff = -1

        if self.packet_type_id_enum in {406, 408, 415, 417}:
            resource_type = ResourceType.ENERGY_REACTIVE
        else:
            resource_type = ResourceType.ENERGY_ACTIVE

        if self.packet_type_id_enum in {400, 401, 402, 403, 404, 405, 406, 407, 408}:
            journal_data_type_for_end = JournalDataType.END_OF_DAY
            period_ago = relativedelta(days=self.period_ago)
            delta = relativedelta(days=1)
            dt_value_current = days_ago_calculation(received_at, device_tz, time(0), period_ago) if self.period_ago != 0 else received_at
            dt_value_previous_1_delta = days_ago_calculation(received_at, device_tz, time(0), period_ago + delta)
            dt_value_previous_2_delta = days_ago_calculation(received_at, device_tz, time(0), period_ago + delta * 2)
            dt_value_previous_3_delta = days_ago_calculation(received_at, device_tz, time(0), period_ago + delta * 3)
        else:
            journal_data_type_for_end = JournalDataType.END_OF_MONTH
            period_ago = relativedelta(months=self.period_ago)
            delta = relativedelta(months=1)
            dt_value_current = first_day_of_month(days_ago_calculation(received_at, device_tz, time(0), period_ago - delta), device_tz) if self.period_ago != 0 else received_at
            dt_value_previous_1_delta = first_day_of_month(days_ago_calculation(received_at, device_tz, time(0), period_ago), device_tz)
            dt_value_previous_2_delta = first_day_of_month(days_ago_calculation(received_at, device_tz, time(0), period_ago + delta), device_tz)
            dt_value_previous_3_delta = first_day_of_month(days_ago_calculation(received_at, device_tz, time(0), period_ago + delta * 2), device_tz)

        integration_message_data = []
        # adding value_current with checking validity
        if self.is_valid:
            integration_message_data.append(
                IntegrationV0MessageData(
                    is_valid=self.is_valid,
                    dt=dt_value_current,
                    consumption=[
                        IntegrationV0MessageConsumption(
                            tariff=tariff,
                            counter_type=CounterType.COMMON,
                            value=Decimal(str(self.value_current)),
                            resource_type=resource_type,
                            channel=1,
                            overloading_value=Decimal(str(1342177.27)),
                            journal_data_type=journal_data_type_for_end if self.period_ago != 0 else JournalDataType.CURRENT,
                        ),
                    ] if self.packet_type_id_enum not in {407, 408, 416, 417} else [],
                    generation=[
                        IntegrationV0MessageGeneration(
                            tariff=tariff,
                            counter_type=CounterType.COMMON,
                            value=Decimal(str(self.value_current)),
                            resource_type=resource_type,
                            overloading_value=Decimal(str(1342177.27)),
                            journal_data_type=journal_data_type_for_end if self.period_ago != 0 else JournalDataType.CURRENT,
                        ),
                    ] if self.packet_type_id_enum in {407, 408, 416, 417} else [],
                ))
            # adding END_OF_DAY / END_OF_MONTH values with checking validity
            integration_message_data.append(
                IntegrationV0MessageData(
                    is_valid=self.is_valid,
                    dt=dt_value_previous_1_delta,
                    consumption=[
                        IntegrationV0MessageConsumption(
                            tariff=tariff,
                            counter_type=CounterType.COMMON,
                            value=Decimal(str(self.value_current - self.value_previous_1_delta)),
                            resource_type=resource_type,
                            channel=1,
                            overloading_value=Decimal(str(167772.15)),
                            journal_data_type=journal_data_type_for_end,
                        ),
                    ] if self.packet_type_id_enum not in {407, 408, 416, 417} else [],
                    generation=[
                        IntegrationV0MessageGeneration(
                            tariff=tariff,
                            counter_type=CounterType.COMMON,
                            value=Decimal(str(self.value_current - self.value_previous_1_delta)),
                            resource_type=resource_type,
                            overloading_value=Decimal(str(167772.15)),
                            journal_data_type=journal_data_type_for_end,
                        ),
                    ] if self.packet_type_id_enum in {407, 408, 416, 417} else [],
                ))
            integration_message_data.append(
                IntegrationV0MessageData(
                    is_valid=self.is_valid,
                    dt=dt_value_previous_2_delta,
                    consumption=[
                        IntegrationV0MessageConsumption(
                            tariff=tariff,
                            counter_type=CounterType.COMMON,
                            value=Decimal(str(self.value_current - self.value_previous_2_delta - self.value_previous_1_delta)),
                            resource_type=resource_type,
                            channel=1,
                            overloading_value=Decimal(str(167772.15)),
                            journal_data_type=journal_data_type_for_end,
                        ),
                    ] if self.packet_type_id_enum not in {407, 408, 416, 417} else [],
                    generation=[
                        IntegrationV0MessageGeneration(
                            tariff=tariff,
                            counter_type=CounterType.COMMON,
                            value=Decimal(str(self.value_current - self.value_previous_2_delta - self.value_previous_1_delta)),
                            resource_type=resource_type,
                            overloading_value=Decimal(str(167772.15)),
                            journal_data_type=journal_data_type_for_end,
                        ),
                    ] if self.packet_type_id_enum in {407, 408, 416, 417} else [],
                ))
            integration_message_data.append(
                IntegrationV0MessageData(
                    is_valid=self.is_valid,
                    dt=dt_value_previous_3_delta,
                    consumption=[
                        IntegrationV0MessageConsumption(
                            tariff=tariff,
                            counter_type=CounterType.COMMON,
                            value=Decimal(str(self.value_current - self.value_previous_3_delta - self.value_previous_2_delta - self.value_previous_1_delta)),
                            resource_type=resource_type,
                            channel=1,
                            overloading_value=Decimal(str(167772.15)),
                            journal_data_type=journal_data_type_for_end,
                        ),
                    ] if self.packet_type_id_enum not in {407, 408, 416, 417} else [],
                    generation=[
                        IntegrationV0MessageGeneration(
                            tariff=tariff,
                            counter_type=CounterType.COMMON,
                            value=Decimal(str(self.value_current - self.value_previous_3_delta - self.value_previous_2_delta - self.value_previous_1_delta)),
                            resource_type=resource_type,
                            overloading_value=Decimal(str(167772.15)),
                            journal_data_type=journal_data_type_for_end,
                        ),
                    ] if self.packet_type_id_enum in {407, 408, 416, 417} else [],
                ))
        return integration_message_data



from typing import Optional


class DataGatewayError(Exception):
    def __init__(self, message: str, error: Optional[Exception] = None) -> None:
        super().__init__(f'{message}. {error}' if error else message)
        self.error = error


class DataGatewayDecryptError(DataGatewayError):
    pass


class DataGatewayDeviceProtocolParsingError(DataGatewayError):
    pass


class DataGatewayBSProtocolParsingError(DataGatewayError):
    pass



import random
import typing
from enum import IntEnum, unique

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_dl_device_energy_16b_get_data
#
# RESULT int:        4294343595594875952597107073
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00001101 11100000 00110010 10100000 11001001 10000011 00111110 00001100 11001011 11000101 00000001 10000001   LSB
# RESULT hex:  LE    81 01 c5 cb 0c 3e 83 c9 a0 32 e0 0d 00 00 00 00
#
# name                             type                                          size  value(int)                                                                                                                        data(bits)
# -------------------------------  --------------------------------------------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL             u7                                               7           1                                                                                                                           0000001
# packet_type_id.0.DFF             bool                                             1           1                                                                                                                          1
# packet_type_id.1.VAL             u2                                               2           1                                                                                                                        01
# packet_type_id.1.DFF             bool                                             1           0                                                                                                                       0
# year                             u7                                               7          32                                                                                                                0100000
# month                            SmpmDlDeviceEnergy16bGetDataDataRequestMonth     4           1                                                                                                            0001
# day                              u5                                               5          15                                                                                                       01111
# request_data_pack_ids.0.pack_id  SmpmDlDeviceEnergy16bGetDataDataRequestId       14         409                                                                                         00000110011001
# request_data_pack_ids.1.pack_id  SmpmDlDeviceEnergy16bGetDataDataRequestId       14         415                                                                           00000110011111
# request_data_pack_ids.2.pack_id  SmpmDlDeviceEnergy16bGetDataDataRequestId       14         403                                                             00000110010011
# request_data_pack_ids.3.pack_id  SmpmDlDeviceEnergy16bGetDataDataRequestId       14         405                                               00000110010101
# request_data_pack_ids.4.pack_id  SmpmDlDeviceEnergy16bGetDataDataRequestId       14         444                                 00000110111100
# request_data_pack_ids.5.pack_id  SmpmDlDeviceEnergy16bGetDataDataRequestId       14           0                   00000000000000
# RESERVED                         u17                                             17           0  00000000000000000
# Reserved space is filled with random bits in DOWNLINK packets !


@unique
class SmpmDlDeviceEnergy16bGetDataDataRequestId(IntEnum):
    UNDEFINED = 0
    UL_DATA_16B__ENERGY = 315
    NETWORK_PARAMS_PHASE1 = 444
    DAILY_ENERGY_ACTIVE_CONSUMED = 400
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_1 = 401
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_2 = 402
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_3 = 403
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_4 = 404
    DAILY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM = 405
    DAILY_ENERGY_REACTIVE_CONSUMED = 406
    DAILY_ENERGY_ACTIVE_GENERATED = 407
    DAILY_ENERGY_REACTIVE_GENERATED = 408
    MONTHLY_ENERGY_ACTIVE_CONSUMED = 409
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_1 = 410
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_2 = 411
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_3 = 412
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_4 = 413
    MONTHLY_ENERGY_ACTIVE_CONSUMED_TARIFF_SUM = 414
    MONTHLY_ENERGY_REACTIVE_CONSUMED = 415
    MONTHLY_ENERGY_ACTIVE_GENERATED = 416
    MONTHLY_ENERGY_REACTIVE_GENERATED = 417


@unique
class SmpmDlDeviceEnergy16bGetDataDataRequestMonth(IntEnum):
    JAN = 1
    FEB = 2
    MAR = 3
    APR = 4
    MAY = 5
    JUN = 6
    JUL = 7
    AUG = 8
    SEP = 9
    OKT = 10
    NOV = 11
    DEC = 12


class SmpmDlDeviceEnergy16BGetDataData(Packet):
    year: int
    month: SmpmDlDeviceEnergy16bGetDataDataRequestMonth
    day: int
    request_data_pack_ids: typing.Tuple[SmpmDlDeviceEnergy16bGetDataDataRequestId, SmpmDlDeviceEnergy16bGetDataDataRequestId, SmpmDlDeviceEnergy16bGetDataDataRequestId, SmpmDlDeviceEnergy16bGetDataDataRequestId, SmpmDlDeviceEnergy16bGetDataDataRequestId, SmpmDlDeviceEnergy16bGetDataDataRequestId]

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((1) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.year, int)
        assert 2000 <= data.year <= 2127
        result |= (((data.year + -2000)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.month, SmpmDlDeviceEnergy16bGetDataDataRequestMonth)
        result |= ((data.month.value) & (2 ** (4) - 1)) << size
        size += 4
        assert isinstance(data.day, int)
        assert 0 <= data.day <= 31
        result |= ((data.day) & (2 ** (5) - 1)) << size
        size += 5
        assert isinstance(data.request_data_pack_ids, tuple) and len(data.request_data_pack_ids) == 6
        assert isinstance(data.request_data_pack_ids[0], SmpmDlDeviceEnergy16bGetDataDataRequestId)
        result |= ((data.request_data_pack_ids[0].value) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.request_data_pack_ids[1], SmpmDlDeviceEnergy16bGetDataDataRequestId)
        result |= ((data.request_data_pack_ids[1].value) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.request_data_pack_ids[2], SmpmDlDeviceEnergy16bGetDataDataRequestId)
        result |= ((data.request_data_pack_ids[2].value) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.request_data_pack_ids[3], SmpmDlDeviceEnergy16bGetDataDataRequestId)
        result |= ((data.request_data_pack_ids[3].value) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.request_data_pack_ids[4], SmpmDlDeviceEnergy16bGetDataDataRequestId)
        result |= ((data.request_data_pack_ids[4].value) & (2 ** (14) - 1)) << size
        size += 14
        assert isinstance(data.request_data_pack_ids[5], SmpmDlDeviceEnergy16bGetDataDataRequestId)
        result |= ((data.request_data_pack_ids[5].value) & (2 ** (14) - 1)) << size
        size += 14
        result |= ((random.getrandbits(17)) & (2 ** 17 - 1)) << size
        size += 17
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmDlDeviceEnergy16BGetDataData':
        result__el_tmp1: typing.Dict[str, typing.Any] = dict()
        if 1 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["year"] = buf.shift(7) + 2000
        result__el_tmp1["month"] = SmpmDlDeviceEnergy16bGetDataDataRequestMonth(buf.shift(4))
        result__el_tmp1["day"] = buf.shift(5) + 0
        request_data_pack_ids_tmp2: typing.List[SmpmDlDeviceEnergy16bGetDataDataRequestId] = []
        request_data_pack_ids__item_tmp3 = SmpmDlDeviceEnergy16bGetDataDataRequestId(buf.shift(14))
        request_data_pack_ids_tmp2.append(request_data_pack_ids__item_tmp3)
        request_data_pack_ids__item_tmp3 = SmpmDlDeviceEnergy16bGetDataDataRequestId(buf.shift(14))
        request_data_pack_ids_tmp2.append(request_data_pack_ids__item_tmp3)
        request_data_pack_ids__item_tmp3 = SmpmDlDeviceEnergy16bGetDataDataRequestId(buf.shift(14))
        request_data_pack_ids_tmp2.append(request_data_pack_ids__item_tmp3)
        request_data_pack_ids__item_tmp3 = SmpmDlDeviceEnergy16bGetDataDataRequestId(buf.shift(14))
        request_data_pack_ids_tmp2.append(request_data_pack_ids__item_tmp3)
        request_data_pack_ids__item_tmp3 = SmpmDlDeviceEnergy16bGetDataDataRequestId(buf.shift(14))
        request_data_pack_ids_tmp2.append(request_data_pack_ids__item_tmp3)
        request_data_pack_ids__item_tmp3 = SmpmDlDeviceEnergy16bGetDataDataRequestId(buf.shift(14))
        request_data_pack_ids_tmp2.append(request_data_pack_ids__item_tmp3)
        result__el_tmp1["request_data_pack_ids"] = tuple(request_data_pack_ids_tmp2)
        result = SmpmDlDeviceEnergy16BGetDataData(**result__el_tmp1)
        buf.shift(17)
        return result



from decimal import Decimal
from datetime import timedelta, datetime, time, tzinfo
from typing import List, Any, Dict
from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, DeviceHack, JournalDataType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageConsumption, \
    IntegrationV0MessageGeneration, CounterType, ResourceType

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_energy_16b_daily
#
# RESULT int:        14738343083975057021112790317536029371
# RESULT bin:  MSB   00001011000101101000000001101010101010101011101100101010101010101010101001010101010111110101010010101110101010101010101010111011   LSB                                 # noqa: E501
# RESULT hex:  LE    bb aa aa ae 54 5f 55 aa aa 2a bb aa 6a 80 16 0b
#
#
# name                       type       size  value(int)                                                                                                                        data(bits)  # noqa: E501
# -------------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------  # noqa: E501
# packet_type_id.0.VAL       u7            7          59                                                                                                                           0111011  # noqa: E501
# packet_type_id.0.DFF       bool          1           1                                                                                                                          1         # noqa: E501
# packet_type_id.1.VAL       u2            2           2                                                                                                                        10          # noqa: E501
# packet_type_id.1.DFF       bool          1           0                                                                                                                       0            # noqa: E501
# energy_consumed_active     u23          23     1430869                                                                                                00101011101010101010101             # noqa: E501
# energy_consumed_reactive   u23          23     1398741                                                                         00101010101011111010101                                    # noqa: E501
# energy_generated_active    u23          23     1398101                                                  00101010101010101010101                                                           # noqa: E501
# energy_generated_reactive  u23          23     6990523                           11010101010101010111011
# days_ago                   timedelta     7           0                    0000000
# valid                      bool          1           0                   0
# error_measurement          bool          1           1                  1
# error_low_voltage          bool          1           0                 0
# error_internal_clock       bool          1           1                1
# error_flash                bool          1           1               1
# error_eeprom               bool          1           0              0
# error_radio                bool          1           1             1
# error_display              bool          1           0            0
# error_plc                  bool          1           0           0
# error_reset                bool          1           0          0
# impact_power_lost          bool          1           1         1
# impact_magnet              bool          1           1        1
# impact_cleat_tamper        bool          1           0       0
# impact_body_tamper         bool          1           1      1
# impact_radio               bool          1           0     0
# RESERVED                   u3            3           0  000


class SmpmUlDeviceEnergy16BDailyData(Packet):
    energy_consumed_active: int
    energy_consumed_reactive: int
    energy_generated_active: int
    energy_generated_reactive: int
    days_ago: timedelta
    valid: bool
    error_measurement: bool
    error_low_voltage: bool
    error_internal_clock: bool
    error_flash: bool
    error_eeprom: bool
    error_radio: bool
    error_display: bool
    error_plc: bool
    error_reset: bool
    impact_power_lost: bool
    impact_magnet: bool
    impact_cleat_tamper: bool
    impact_body_tamper: bool
    impact_radio: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((59) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((2) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.energy_consumed_active, int)
        result |= (((data.energy_consumed_active) & 8388607) & (2 ** (23) - 1)) << size
        size += 23
        assert isinstance(data.energy_consumed_reactive, int)
        result |= (((data.energy_consumed_reactive) & 8388607) & (2 ** (23) - 1)) << size
        size += 23
        assert isinstance(data.energy_generated_active, int)
        result |= (((data.energy_generated_active) & 8388607) & (2 ** (23) - 1)) << size
        size += 23
        assert isinstance(data.energy_generated_reactive, int)
        result |= (((data.energy_generated_reactive) & 8388607) & (2 ** (23) - 1)) << size
        size += 23
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 127
        result |= ((days_ago_tmp1) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.valid, bool)
        result |= ((int(data.valid)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_measurement, bool)
        result |= ((int(data.error_measurement)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_low_voltage, bool)
        result |= ((int(data.error_low_voltage)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_internal_clock, bool)
        result |= ((int(data.error_internal_clock)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_flash, bool)
        result |= ((int(data.error_flash)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_eeprom, bool)
        result |= ((int(data.error_eeprom)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_radio, bool)
        result |= ((int(data.error_radio)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_display, bool)
        result |= ((int(data.error_display)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_plc, bool)
        result |= ((int(data.error_plc)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.error_reset, bool)
        result |= ((int(data.error_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.impact_power_lost, bool)
        result |= ((int(data.impact_power_lost)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.impact_magnet, bool)
        result |= ((int(data.impact_magnet)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.impact_cleat_tamper, bool)
        result |= ((int(data.impact_cleat_tamper)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.impact_body_tamper, bool)
        result |= ((int(data.impact_body_tamper)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.impact_radio, bool)
        result |= ((int(data.impact_radio)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BDailyData':
        result__el_tmp2: Dict[str, Any] = dict()
        if 59 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 2 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp2["energy_consumed_active"] = buf.shift(23) + 0
        result__el_tmp2["energy_consumed_reactive"] = buf.shift(23) + 0
        result__el_tmp2["energy_generated_active"] = buf.shift(23) + 0
        result__el_tmp2["energy_generated_reactive"] = buf.shift(23) + 0
        result__el_tmp2["days_ago"] = timedelta(seconds=buf.shift(7) * 86400)
        result__el_tmp2["valid"] = bool(buf.shift(1))
        result__el_tmp2["error_measurement"] = bool(buf.shift(1))
        result__el_tmp2["error_low_voltage"] = bool(buf.shift(1))
        result__el_tmp2["error_internal_clock"] = bool(buf.shift(1))
        result__el_tmp2["error_flash"] = bool(buf.shift(1))
        result__el_tmp2["error_eeprom"] = bool(buf.shift(1))
        result__el_tmp2["error_radio"] = bool(buf.shift(1))
        result__el_tmp2["error_display"] = bool(buf.shift(1))
        result__el_tmp2["error_plc"] = bool(buf.shift(1))
        result__el_tmp2["error_reset"] = bool(buf.shift(1))
        result__el_tmp2["impact_power_lost"] = bool(buf.shift(1))
        result__el_tmp2["impact_magnet"] = bool(buf.shift(1))
        result__el_tmp2["impact_cleat_tamper"] = bool(buf.shift(1))
        result__el_tmp2["impact_body_tamper"] = bool(buf.shift(1))
        result__el_tmp2["impact_radio"] = bool(buf.shift(1))
        result = SmpmUlDeviceEnergy16BDailyData(**result__el_tmp2)
        buf.shift(3)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        overloading_value = 8388607.0
        hacks = kwargs.get('hacks', [])
        if DeviceHack.electricity_daily_packet_overload_value in hacks:
            overloading_value = 838861.0
        return [
            IntegrationV0MessageData(
                dt=days_ago_calculation(received_at, device_tz, time(0), self.days_ago),
                is_valid=self.valid,
                consumption=[
                    IntegrationV0MessageConsumption(
                        value=Decimal(str(self.energy_consumed_active)),
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.ENERGY_ACTIVE,
                        channel=1,
                        overloading_value=Decimal(str(overloading_value)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageConsumption(
                        value=Decimal(str(self.energy_consumed_reactive)),
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.ENERGY_REACTIVE,
                        channel=1,
                        overloading_value=Decimal(str(overloading_value)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.ERROR_MEASUREMENT] if self.error_measurement else []),
                    *([IntegrationV0MessageEvent.ERROR_LOW_VOLTAGE] if self.error_low_voltage else []),
                    *([IntegrationV0MessageEvent.ERROR_INTERNAL_CLOCK] if self.error_internal_clock else []),
                    *([IntegrationV0MessageEvent.ERROR_FLASH] if self.error_flash else []),
                    *([IntegrationV0MessageEvent.ERROR_EEPROM] if self.error_eeprom else []),
                    *([IntegrationV0MessageEvent.ERROR_RADIO] if self.error_radio else []),
                    *([IntegrationV0MessageEvent.ERROR_DISPLAY] if self.error_display else []),
                    *([IntegrationV0MessageEvent.ERROR_PLC] if self.error_plc else []),
                    *([IntegrationV0MessageEvent.ERROR_RESET] if self.error_reset else []),
                    *([IntegrationV0MessageEvent.IMPACT_POWER_LOST] if self.impact_power_lost else []),
                    *([IntegrationV0MessageEvent.IMPACT_MAGNET] if self.impact_magnet else []),
                    *([IntegrationV0MessageEvent.IMPACT_CLEAT_TAMPER] if self.impact_cleat_tamper else []),
                    *([IntegrationV0MessageEvent.IMPACT_RADIO] if self.impact_radio else []),
                ],
                generation=[
                    IntegrationV0MessageGeneration(
                        value=Decimal(str(self.energy_generated_active)),
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.ENERGY_ACTIVE,
                        overloading_value=Decimal(str(overloading_value)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                    IntegrationV0MessageGeneration(
                        value=Decimal(str(self.energy_generated_reactive)),
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.ENERGY_REACTIVE,
                        overloading_value=Decimal(str(overloading_value)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
            ),
        ]



from datetime import timedelta, datetime, time, tzinfo
from enum import IntEnum, unique
from typing import List, Any, Dict, Tuple, Optional

from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageProfile, ProfileGranulation, GRANULATION_TO_END_OF_DATETIME_MAP, ProfileKind
from data_aggregator_sdk.utils.round_dt import round_dt

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet
from data_gateway_sdk.utils.true_round import true_round


# PACKET (128 bits)   smpm_ul_device_energy_16b_profile_8h_pqs
#
# RESULT int:        16227037341613019245223190382775473
# RESULT bin:  MSB   00000000 00000011 00100000 00001110 00000001 10000000 00101000 00000100 00000000 01100000 00001000 00000000 10000000 00000000 00001100 10110001   LSB
# RESULT hex:  LE    b1 0c 00 80 00 08 60 00 04 28 80 01 0e 20 03 00
#
# name                  type       size  value(int)                                                                                                                        data(bits)   # noqa: E501
# --------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL  u7            7          49                                                                                                                           0110001   # noqa: E501
# packet_type_id.0.DFF  bool          1           1                                                                                                                          1
# packet_type_id.1.VAL  u2            2           0                                                                                                                        00
# packet_type_id.1.DFF  bool          1           1                                                                                                                       1
# packet_type_id.2.VAL  u2            2           1                                                                                                                     01
# packet_type_id.2.DFF  bool          1           0                                                                                                                    0
# days_ago              timedelta     6           0                                                                                                              000000
# profile.0.point       u11          11           0                                                                                                   00000000000
# profile.1.point       u11          11           1                                                                                        00000000001
# profile.2.point       u11          11           2                                                                             00000000010
# profile.3.point       u11          11           3                                                                  00000000011
# profile.4.point       u11          11           4                                                       00000000100
# profile.5.point       u11          11           5                                            00000000101
# profile.6.point       u11          11           6                                 00000000110
# profile.7.point       u11          11           7                      00000000111
# point_factor          uf12p1       12          50          000000110010
# RESERVED              u8            8           0  00000000

@unique
class SmpmUlDeviceEnergy16bProfile8hPQSIds(IntEnum):
    UL_DATA_16B__PROFILE_H8_1_S__AVG = 561
    UL_DATA_16B__PROFILE_H8_1_S__AVG__PHASE_A = 562
    UL_DATA_16B__PROFILE_H8_1_S__AVG__PHASE_B = 563
    UL_DATA_16B__PROFILE_H8_1_S__AVG__PHASE_C = 564
    UL_DATA_16B__PROFILE_H8_1_S__MIN = 565
    UL_DATA_16B__PROFILE_H8_1_S__MIN__PHASE_A = 566
    UL_DATA_16B__PROFILE_H8_1_S__MIN__PHASE_B = 567
    UL_DATA_16B__PROFILE_H8_1_S__MIN__PHASE_C = 568
    UL_DATA_16B__PROFILE_H8_1_S__MAX = 569
    UL_DATA_16B__PROFILE_H8_1_S__MAX__PHASE_A = 570
    UL_DATA_16B__PROFILE_H8_1_S__MAX__PHASE_B = 571
    UL_DATA_16B__PROFILE_H8_1_S__MAX__PHASE_C = 572
    UL_DATA_16B__PROFILE_H8_1_P__AVG = 573
    UL_DATA_16B__PROFILE_H8_1_P__AVG__PHASE_A = 574
    UL_DATA_16B__PROFILE_H8_1_P__AVG__PHASE_B = 575
    UL_DATA_16B__PROFILE_H8_1_P__AVG__PHASE_C = 576
    UL_DATA_16B__PROFILE_H8_1_P__MIN = 577
    UL_DATA_16B__PROFILE_H8_1_P__MIN__PHASE_A = 578
    UL_DATA_16B__PROFILE_H8_1_P__MIN__PHASE_B = 579
    UL_DATA_16B__PROFILE_H8_1_P__MIN__PHASE_C = 580
    UL_DATA_16B__PROFILE_H8_1_P__MAX = 581
    UL_DATA_16B__PROFILE_H8_1_P__MAX__PHASE_A = 582
    UL_DATA_16B__PROFILE_H8_1_P__MAX__PHASE_B = 583
    UL_DATA_16B__PROFILE_H8_1_P__MAX__PHASE_C = 584
    UL_DATA_16B__PROFILE_H8_1_Q__AVG = 585
    UL_DATA_16B__PROFILE_H8_1_Q__AVG__PHASE_A = 586
    UL_DATA_16B__PROFILE_H8_1_Q__AVG__PHASE_B = 587
    UL_DATA_16B__PROFILE_H8_1_Q__AVG__PHASE_C = 588
    UL_DATA_16B__PROFILE_H8_1_Q__MIN = 589
    UL_DATA_16B__PROFILE_H8_1_Q__MIN__PHASE_A = 590
    UL_DATA_16B__PROFILE_H8_1_Q__MIN__PHASE_B = 591
    UL_DATA_16B__PROFILE_H8_1_Q__MIN__PHASE_C = 592
    UL_DATA_16B__PROFILE_H8_1_Q__MAX = 593
    UL_DATA_16B__PROFILE_H8_1_Q__MAX__PHASE_A = 594
    UL_DATA_16B__PROFILE_H8_1_Q__MAX__PHASE_B = 595
    UL_DATA_16B__PROFILE_H8_1_Q__MAX__PHASE_C = 596
    UL_DATA_16B__PROFILE_H8_2_S__AVG = 597
    UL_DATA_16B__PROFILE_H8_2_S__AVG__PHASE_A = 598
    UL_DATA_16B__PROFILE_H8_2_S__AVG__PHASE_B = 599
    UL_DATA_16B__PROFILE_H8_2_S__AVG__PHASE_C = 600
    UL_DATA_16B__PROFILE_H8_2_S__MIN = 601
    UL_DATA_16B__PROFILE_H8_2_S__MIN__PHASE_A = 602
    UL_DATA_16B__PROFILE_H8_2_S__MIN__PHASE_B = 603
    UL_DATA_16B__PROFILE_H8_2_S__MIN__PHASE_C = 604
    UL_DATA_16B__PROFILE_H8_2_S__MAX = 605
    UL_DATA_16B__PROFILE_H8_2_S__MAX__PHASE_A = 606
    UL_DATA_16B__PROFILE_H8_2_S__MAX__PHASE_B = 607
    UL_DATA_16B__PROFILE_H8_2_S__MAX__PHASE_C = 608
    UL_DATA_16B__PROFILE_H8_2_P__AVG = 609
    UL_DATA_16B__PROFILE_H8_2_P__AVG__PHASE_A = 610
    UL_DATA_16B__PROFILE_H8_2_P__AVG__PHASE_B = 611
    UL_DATA_16B__PROFILE_H8_2_P__AVG__PHASE_C = 612
    UL_DATA_16B__PROFILE_H8_2_P__MIN = 613
    UL_DATA_16B__PROFILE_H8_2_P__MIN__PHASE_A = 614
    UL_DATA_16B__PROFILE_H8_2_P__MIN__PHASE_B = 615
    UL_DATA_16B__PROFILE_H8_2_P__MIN__PHASE_C = 616
    UL_DATA_16B__PROFILE_H8_2_P__MAX = 617
    UL_DATA_16B__PROFILE_H8_2_P__MAX__PHASE_A = 618
    UL_DATA_16B__PROFILE_H8_2_P__MAX__PHASE_B = 619
    UL_DATA_16B__PROFILE_H8_2_P__MAX__PHASE_C = 620
    UL_DATA_16B__PROFILE_H8_2_Q__AVG = 621
    UL_DATA_16B__PROFILE_H8_2_Q__AVG__PHASE_A = 622
    UL_DATA_16B__PROFILE_H8_2_Q__AVG__PHASE_B = 623
    UL_DATA_16B__PROFILE_H8_2_Q__AVG__PHASE_C = 624
    UL_DATA_16B__PROFILE_H8_2_Q__MIN = 625
    UL_DATA_16B__PROFILE_H8_2_Q__MIN__PHASE_A = 626
    UL_DATA_16B__PROFILE_H8_2_Q__MIN__PHASE_B = 627
    UL_DATA_16B__PROFILE_H8_2_Q__MIN__PHASE_C = 628
    UL_DATA_16B__PROFILE_H8_2_Q__MAX = 629
    UL_DATA_16B__PROFILE_H8_2_Q__MAX__PHASE_A = 630
    UL_DATA_16B__PROFILE_H8_2_Q__MAX__PHASE_B = 631
    UL_DATA_16B__PROFILE_H8_2_Q__MAX__PHASE_C = 632
    UL_DATA_16B__PROFILE_H8_3_S__AVG = 633
    UL_DATA_16B__PROFILE_H8_3_S__AVG__PHASE_A = 634
    UL_DATA_16B__PROFILE_H8_3_S__AVG__PHASE_B = 635
    UL_DATA_16B__PROFILE_H8_3_S__AVG__PHASE_C = 636
    UL_DATA_16B__PROFILE_H8_3_S__MIN = 637
    UL_DATA_16B__PROFILE_H8_3_S__MIN__PHASE_A = 638
    UL_DATA_16B__PROFILE_H8_3_S__MIN__PHASE_B = 639
    UL_DATA_16B__PROFILE_H8_3_S__MIN__PHASE_C = 640
    UL_DATA_16B__PROFILE_H8_3_S__MAX = 641
    UL_DATA_16B__PROFILE_H8_3_S__MAX__PHASE_A = 642
    UL_DATA_16B__PROFILE_H8_3_S__MAX__PHASE_B = 643
    UL_DATA_16B__PROFILE_H8_3_S__MAX__PHASE_C = 644
    UL_DATA_16B__PROFILE_H8_3_P__AVG = 645
    UL_DATA_16B__PROFILE_H8_3_P__AVG__PHASE_A = 646
    UL_DATA_16B__PROFILE_H8_3_P__AVG__PHASE_B = 647
    UL_DATA_16B__PROFILE_H8_3_P__AVG__PHASE_C = 648
    UL_DATA_16B__PROFILE_H8_3_P__MIN = 649
    UL_DATA_16B__PROFILE_H8_3_P__MIN__PHASE_A = 650
    UL_DATA_16B__PROFILE_H8_3_P__MIN__PHASE_B = 651
    UL_DATA_16B__PROFILE_H8_3_P__MIN__PHASE_C = 652
    UL_DATA_16B__PROFILE_H8_3_P__MAX = 653
    UL_DATA_16B__PROFILE_H8_3_P__MAX__PHASE_A = 654
    UL_DATA_16B__PROFILE_H8_3_P__MAX__PHASE_B = 655
    UL_DATA_16B__PROFILE_H8_3_P__MAX__PHASE_C = 656
    UL_DATA_16B__PROFILE_H8_3_Q__AVG = 657
    UL_DATA_16B__PROFILE_H8_3_Q__AVG__PHASE_A = 658
    UL_DATA_16B__PROFILE_H8_3_Q__AVG__PHASE_B = 659
    UL_DATA_16B__PROFILE_H8_3_Q__AVG__PHASE_C = 660
    UL_DATA_16B__PROFILE_H8_3_Q__MIN = 661
    UL_DATA_16B__PROFILE_H8_3_Q__MIN__PHASE_A = 662
    UL_DATA_16B__PROFILE_H8_3_Q__MIN__PHASE_B = 663
    UL_DATA_16B__PROFILE_H8_3_Q__MIN__PHASE_C = 664
    UL_DATA_16B__PROFILE_H8_3_Q__MAX = 665
    UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_A = 666
    UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_B = 667
    UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C = 668


PROFILE_H8_PQS_TYPE_MAP = {
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG: ProfileKind.FULL_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG__PHASE_A: ProfileKind.FULL_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG__PHASE_B: ProfileKind.FULL_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__AVG__PHASE_C: ProfileKind.FULL_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__MIN: ProfileKind.FULL_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__MIN__PHASE_A: ProfileKind.FULL_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__MIN__PHASE_B: ProfileKind.FULL_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__MIN__PHASE_C: ProfileKind.FULL_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__MAX: ProfileKind.FULL_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__MAX__PHASE_A: ProfileKind.FULL_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__MAX__PHASE_B: ProfileKind.FULL_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_S__MAX__PHASE_C: ProfileKind.FULL_POWER_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__AVG: ProfileKind.ACTIVE_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__AVG__PHASE_A: ProfileKind.ACTIVE_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__AVG__PHASE_B: ProfileKind.ACTIVE_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__AVG__PHASE_C: ProfileKind.ACTIVE_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__MIN: ProfileKind.ACTIVE_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__MIN__PHASE_A: ProfileKind.ACTIVE_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__MIN__PHASE_B: ProfileKind.ACTIVE_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__MIN__PHASE_C: ProfileKind.ACTIVE_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__MAX: ProfileKind.ACTIVE_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__MAX__PHASE_A: ProfileKind.ACTIVE_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__MAX__PHASE_B: ProfileKind.ACTIVE_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_P__MAX__PHASE_C: ProfileKind.ACTIVE_POWER_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__AVG: ProfileKind.REACTIVE_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__AVG__PHASE_A: ProfileKind.REACTIVE_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__AVG__PHASE_B: ProfileKind.REACTIVE_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__AVG__PHASE_C: ProfileKind.REACTIVE_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__MIN: ProfileKind.REACTIVE_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__MIN__PHASE_A: ProfileKind.REACTIVE_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__MIN__PHASE_B: ProfileKind.REACTIVE_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__MIN__PHASE_C: ProfileKind.REACTIVE_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__MAX: ProfileKind.REACTIVE_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__MAX__PHASE_A: ProfileKind.REACTIVE_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__MAX__PHASE_B: ProfileKind.REACTIVE_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_1_Q__MAX__PHASE_C: ProfileKind.REACTIVE_POWER_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__AVG: ProfileKind.FULL_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__AVG__PHASE_A: ProfileKind.FULL_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__AVG__PHASE_B: ProfileKind.FULL_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__AVG__PHASE_C: ProfileKind.FULL_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__MIN: ProfileKind.FULL_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__MIN__PHASE_A: ProfileKind.FULL_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__MIN__PHASE_B: ProfileKind.FULL_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__MIN__PHASE_C: ProfileKind.FULL_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__MAX: ProfileKind.FULL_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__MAX__PHASE_A: ProfileKind.FULL_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__MAX__PHASE_B: ProfileKind.FULL_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_S__MAX__PHASE_C: ProfileKind.FULL_POWER_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__AVG: ProfileKind.ACTIVE_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__AVG__PHASE_A: ProfileKind.ACTIVE_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__AVG__PHASE_B: ProfileKind.ACTIVE_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__AVG__PHASE_C: ProfileKind.ACTIVE_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__MIN: ProfileKind.ACTIVE_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__MIN__PHASE_A: ProfileKind.ACTIVE_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__MIN__PHASE_B: ProfileKind.ACTIVE_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__MIN__PHASE_C: ProfileKind.ACTIVE_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__MAX: ProfileKind.ACTIVE_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__MAX__PHASE_A: ProfileKind.ACTIVE_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__MAX__PHASE_B: ProfileKind.ACTIVE_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_P__MAX__PHASE_C: ProfileKind.ACTIVE_POWER_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__AVG: ProfileKind.REACTIVE_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__AVG__PHASE_A: ProfileKind.REACTIVE_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__AVG__PHASE_B: ProfileKind.REACTIVE_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__AVG__PHASE_C: ProfileKind.REACTIVE_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__MIN: ProfileKind.REACTIVE_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__MIN__PHASE_A: ProfileKind.REACTIVE_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__MIN__PHASE_B: ProfileKind.REACTIVE_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__MIN__PHASE_C: ProfileKind.REACTIVE_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__MAX: ProfileKind.REACTIVE_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__MAX__PHASE_A: ProfileKind.REACTIVE_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__MAX__PHASE_B: ProfileKind.REACTIVE_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_2_Q__MAX__PHASE_C: ProfileKind.REACTIVE_POWER_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__AVG: ProfileKind.FULL_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__AVG__PHASE_A: ProfileKind.FULL_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__AVG__PHASE_B: ProfileKind.FULL_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__AVG__PHASE_C: ProfileKind.FULL_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__MIN: ProfileKind.FULL_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__MIN__PHASE_A: ProfileKind.FULL_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__MIN__PHASE_B: ProfileKind.FULL_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__MIN__PHASE_C: ProfileKind.FULL_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__MAX: ProfileKind.FULL_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__MAX__PHASE_A: ProfileKind.FULL_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__MAX__PHASE_B: ProfileKind.FULL_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_S__MAX__PHASE_C: ProfileKind.FULL_POWER_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__AVG: ProfileKind.ACTIVE_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__AVG__PHASE_A: ProfileKind.ACTIVE_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__AVG__PHASE_B: ProfileKind.ACTIVE_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__AVG__PHASE_C: ProfileKind.ACTIVE_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__MIN: ProfileKind.ACTIVE_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__MIN__PHASE_A: ProfileKind.ACTIVE_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__MIN__PHASE_B: ProfileKind.ACTIVE_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__MIN__PHASE_C: ProfileKind.ACTIVE_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__MAX: ProfileKind.ACTIVE_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__MAX__PHASE_A: ProfileKind.ACTIVE_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__MAX__PHASE_B: ProfileKind.ACTIVE_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_P__MAX__PHASE_C: ProfileKind.ACTIVE_POWER_MAX_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__AVG: ProfileKind.REACTIVE_POWER_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__AVG__PHASE_A: ProfileKind.REACTIVE_POWER_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__AVG__PHASE_B: ProfileKind.REACTIVE_POWER_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__AVG__PHASE_C: ProfileKind.REACTIVE_POWER_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MIN: ProfileKind.REACTIVE_POWER_MIN_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MIN__PHASE_A: ProfileKind.REACTIVE_POWER_MIN_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MIN__PHASE_B: ProfileKind.REACTIVE_POWER_MIN_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MIN__PHASE_C: ProfileKind.REACTIVE_POWER_MIN_C,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX: ProfileKind.REACTIVE_POWER_MAX_ABC,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_A: ProfileKind.REACTIVE_POWER_MAX_A,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_B: ProfileKind.REACTIVE_POWER_MAX_B,
    SmpmUlDeviceEnergy16bProfile8hPQSIds.UL_DATA_16B__PROFILE_H8_3_Q__MAX__PHASE_C: ProfileKind.REACTIVE_POWER_MAX_C,
}


class SmpmUlDeviceEnergy16BProfile8HPqsData(Packet):
    packet_type_id: SmpmUlDeviceEnergy16bProfile8hPQSIds
    days_ago: timedelta
    profile: Tuple[int, int, int, int, int, int, int, int]
    point_factor: float

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        packet_type_id__value_int_tmp1 = 0
        packet_type_id__value_size_tmp2 = 0
        assert isinstance(data.packet_type_id, SmpmUlDeviceEnergy16bProfile8hPQSIds)
        packet_type_id__value_int_tmp1 |= ((data.packet_type_id.value) & (2 ** (10) - 1)) << packet_type_id__value_size_tmp2
        packet_type_id__value_size_tmp2 += 10
        packet_type_id__max_size_tmp3 = 0
        packet_type_id__steps_tmp4 = [8, 3, 3, 3]
        for packet_type_id__j_tmp5 in range(32):
            packet_type_id__step_tmp6 = packet_type_id__steps_tmp4[packet_type_id__j_tmp5] if packet_type_id__j_tmp5 < len(packet_type_id__steps_tmp4) else packet_type_id__steps_tmp4[-1]  # noqa: E501
            packet_type_id__max_size_tmp3 += packet_type_id__step_tmp6
            packet_type_id__current_part_value_tmp7 = packet_type_id__value_int_tmp1 & (2 ** packet_type_id__step_tmp6 - 1)
            packet_type_id__value_int_tmp1 = packet_type_id__value_int_tmp1 >> (packet_type_id__step_tmp6 - 1)
            result |= ((packet_type_id__current_part_value_tmp7) & (2 ** ((packet_type_id__step_tmp6 - 1)) - 1)) << size
            size += (packet_type_id__step_tmp6 - 1)
            assert isinstance((packet_type_id__value_int_tmp1 != 0), bool)
            result |= ((int((packet_type_id__value_int_tmp1 != 0))) & (2 ** (1) - 1)) << size
            size += 1
            if packet_type_id__value_int_tmp1 == 0:
                break
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp8 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp8 <= 63
        result |= ((days_ago_tmp8) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.profile, tuple) and len(data.profile) == 8
        assert isinstance(data.profile[0], int)
        assert 0 <= data.profile[0] <= 4095
        result |= ((data.profile[0]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[1], int)
        assert 0 <= data.profile[1] <= 4095
        result |= ((data.profile[1]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[2], int)
        assert 0 <= data.profile[2] <= 4095
        result |= ((data.profile[2]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[3], int)
        assert 0 <= data.profile[3] <= 4095
        result |= ((data.profile[3]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[4], int)
        assert 0 <= data.profile[4] <= 4095
        result |= ((data.profile[4]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[5], int)
        assert 0 <= data.profile[5] <= 4095
        result |= ((data.profile[5]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[6], int)
        assert 0 <= data.profile[6] <= 4095
        result |= ((data.profile[6]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[7], int)
        assert 0 <= data.profile[7] <= 4095
        result |= ((data.profile[7]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.point_factor, (int, float))
        assert 0.0 <= data.point_factor <= 40.95
        result |= ((int(round(float(data.point_factor) * 100.0, 0))) & (2 ** (12) - 1)) << size
        size += 12
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BProfile8HPqsData':
        result__el_tmp9: Dict[str, Any] = dict()
        packet_type_id__res_tmp10 = 0
        packet_type_id__steps_tmp11 = (8, 3, 3, 3)
        packet_type_id__res_size_tmp12 = 0
        packet_type_id__step_tmp13 = 0
        for packet_type_id__i_tmp15 in range(32):
            packet_type_id__step_tmp13 = (packet_type_id__steps_tmp11[packet_type_id__i_tmp15] if packet_type_id__i_tmp15 < len(packet_type_id__steps_tmp11) else packet_type_id__steps_tmp11[-1]) - 1  # noqa: E501
            packet_type_id__res_tmp10 |= buf.shift(packet_type_id__step_tmp13) << packet_type_id__res_size_tmp12
            packet_type_id__res_size_tmp12 += packet_type_id__step_tmp13
            packet_type_id__dff_tmp14 = bool(buf.shift(1))
            if not packet_type_id__dff_tmp14:
                break
        packet_type_id__buf_tmp16 = buf
        buf = BufRef(packet_type_id__res_tmp10, stop_on_buffer_end=True)
        result__el_tmp9["packet_type_id"] = SmpmUlDeviceEnergy16bProfile8hPQSIds(buf.shift(10))
        buf = packet_type_id__buf_tmp16
        result__el_tmp9["days_ago"] = timedelta(seconds=buf.shift(6) * 86400)
        profile_tmp17: List[int] = []
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        result__el_tmp9["profile"] = tuple(profile_tmp17)
        result__el_tmp9["point_factor"] = round(buf.shift(12) / 100.0, 2)
        result = SmpmUlDeviceEnergy16BProfile8HPqsData(**result__el_tmp9)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        values: Tuple[Optional[float], ...]
        profiles = (
            true_round(self.profile[0] * self.point_factor) / 1000,
            true_round(self.profile[1] * self.point_factor) / 1000,
            true_round(self.profile[2] * self.point_factor) / 1000,
            true_round(self.profile[3] * self.point_factor) / 1000,
            true_round(self.profile[4] * self.point_factor) / 1000,
            true_round(self.profile[5] * self.point_factor) / 1000,
            true_round(self.profile[6] * self.point_factor) / 1000,
            true_round(self.profile[7] * self.point_factor) / 1000,
        )
        if self.packet_type_id in {561, 562, 563, 564, 565, 566, 567, 568, 569, 570, 571, 572, 573, 574, 575, 576, 577, 578, 579,
                                   580, 581, 582, 583, 584, 585, 586, 587, 588, 589, 590, 591, 592, 593, 594, 595, 596}:
            values = (
                *profiles,
                None, None, None, None, None, None, None, None,
                None, None, None, None, None, None, None, None,
            )
        elif self.packet_type_id in {597, 598, 599, 600, 601, 602, 603, 604, 605, 606, 607, 608, 609, 610, 611, 612, 613, 614, 615,
                                     616, 617, 618, 619, 620, 621, 622, 623, 624, 625, 626, 627, 628, 629, 630, 631, 632}:
            values = (
                None, None, None, None, None, None, None, None,
                *profiles,
                None, None, None, None, None, None, None, None,
            )
        else:
            values = (
                None, None, None, None, None, None, None, None,
                None, None, None, None, None, None, None, None,
                *profiles,
            )
        return [
            IntegrationV0MessageData(
                dt=round_dt(
                    days_ago_calculation(
                        received_at,
                        device_tz,
                        time(0),
                        self.days_ago,
                    ),
                    GRANULATION_TO_END_OF_DATETIME_MAP[ProfileGranulation.MINUTE_60],
                ),
                profiles=[
                    IntegrationV0MessageProfile(
                        type=PROFILE_H8_PQS_TYPE_MAP[self.packet_type_id],
                        granulation=ProfileGranulation.MINUTE_60,
                        values=values,
                    ),
                ],
            ),
        ]
